from chariot.transformer.base_preprocessor import BasePreprocessor


class TextNormalizer(BasePreprocessor):

    def apply(self, text):
        raise Exception("You have to implements apply")


class TextFilter(BasePreprocessor):

    def apply(self, text):
        raise Exception("You have to implements apply")



from sklearn.base import BaseEstimator, TransformerMixin
from chariot.util import apply_map


class BasePreprocessor(BaseEstimator, TransformerMixin):

    def __init__(self, copy=True):
        self.copy = copy

    def apply(self, elements):
        raise Exception("You have to implements apply")

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        return apply_map(X, self.apply, inplace=(not self.copy))



from chariot.transformer.text.base import TextFilter


class SymbolFilter(TextFilter):

    def __init__(self, filters='!"#$%&()*+,-./:;<=>?@[\\]^_`{|}~\t\n',
                 split=" ", copy=True):
        super().__init__(copy)
        self.filters = filters
        self.split = split

    def apply(self, text):
        translate_dict = dict((c, self.split) for c in self.filters)
        translate_map = str.maketrans(translate_dict)
        _text = text.translate(translate_map)

        return _text



import unicodedata
from chariot.transformer.text.base import TextNormalizer


class UnicodeNormalizer(TextNormalizer):

    def __init__(self, form="NFKC", copy=True):
        super().__init__(copy)
        self.form = form

    def apply(self, text):
        return unicodedata.normalize(self.form, text)



from sklearn.base import BaseEstimator, TransformerMixin


class BaseFormatter(BaseEstimator, TransformerMixin):

    def __init__(self):
        pass

    def transform(self, column):
        raise Exception("You have to implements transform method.")



import numpy as np
from chariot.transformer.formatter.base import BaseFormatter
from chariot.preprocessor import Preprocessor


class CategoricalLabel(BaseFormatter):

    def __init__(self, num_class=-1):
        super().__init__()
        self.num_class = num_class

    def transfer_setting(self, vocabulary_or_preprocessor):
        vocabulary = vocabulary_or_preprocessor
        if isinstance(vocabulary_or_preprocessor, Preprocessor):
            vocabulary = vocabulary_or_preprocessor.vocabulary
        self.num_class = vocabulary.count

    def transform(self, column):
        y = np.array(column, dtype="int")
        input_shape = y.shape
        if input_shape and input_shape[-1] == 1 and len(input_shape) > 1:
            input_shape = tuple(input_shape[:-1])
        y = y.ravel()
        n = y.shape[0]
        categorical = np.zeros((n, self.num_class), dtype=np.float32)
        categorical[np.arange(n), y] = 1
        output_shape = input_shape + (self.num_class,)
        categorical = np.reshape(categorical, output_shape)
        return categorical

    def inverse_transform(self, batch):
        return np.argmax(batch, axis=1)



import numpy as np
import pandas as pd
from chariot.transformer.formatter.base import BaseFormatter
from chariot.transformer.generator.base import BaseGenerator
from chariot.preprocessor import Preprocessor
from chariot.base_dataset_preprocessor import BaseDatasetPreprocessor


class ProcessBuilder():

    def __init__(self, dp, name):
        self.dp = dp
        self.key = (name, name)
        if self.key not in self.dp.spec:
            self.dp.spec[self.key] = {}

    @property
    def preprocessor(self):
        if "preprocessor" in self.dp.spec[self.key]:
            return self.dp.spec[self.key]["preprocessor"]
        else:
            return None

    @property
    def formatter(self):
        if "formatter" in self.dp.spec[self.key]:
            return self.dp.spec[self.key]["formatter"]
        else:
            return None

    def fit(self, X, y=None):
        if self.preprocessor is not None:
            self.preprocessor.fit(X, y)
            if self.formatter is not None:
                f = self._convey(self.formatter)
                self.dp.spec[self.key]["formatter"] = f
        return self

    def by(self, processor, reference=None):
        if isinstance(processor, (BaseFormatter, BaseGenerator)):
            p = None if reference is None else reference.preprocessor
            self.dp.spec[self.key]["formatter"] = self._convey(processor, p)
        else:
            if self.preprocessor is None:
                self.dp.spec[self.key]["preprocessor"] = Preprocessor()
            if isinstance(processor, Preprocessor):
                self.dp.spec[self.key]["preprocessor"] = processor
            else:
                self.dp.spec[self.key]["preprocessor"].stack(processor)

        return self

    def as_name(self, to_name):
        new_key = (self.key[0], to_name)
        self.dp.spec[new_key] = self.dp.spec.pop(self.key)
        self.key = new_key
        return ProcessBuilder(self.dp, to_name)

    def _convey(self, formatter, preprocessor=None):
        _p = preprocessor if preprocessor is not None else self.preprocessor
        if _p is not None and hasattr(formatter, "transfer_setting"):
            formatter.transfer_setting(_p)
        return formatter


class DatasetPreprocessor(BaseDatasetPreprocessor):

    def __init__(self, spec=()):
        super().__init__(spec)
        self._processed = None
        self._formatted = False

    @property
    def processed(self):
        return self._processed

    def process(self, name):
        return ProcessBuilder(self, name)

    def get_spec(self, kind):
        spec = {}
        for f in self.spec:
            if kind in self.spec[f]:
                spec[f] = self.spec[f][kind]

        return spec

    def __call__(self, data):
        self._processed = data
        self._formatted = False
        return self

    def preprocess(self, data=None, n_jobs=1, inverse=False,
                   as_dataframe=False):
        spec = self.get_spec("preprocessor")
        _data = data if data is not None else self._processed
        _data = self.transform(spec, _data, n_jobs, inverse, as_dataframe)

        if data is None:
            self._processed = _data
            return self
        else:
            return _data

    def format(self, data=None, n_jobs=1, inverse=False,
               as_dataframe=False):
        spec = self.get_spec("formatter")
        _data = data if data is not None else self._processed
        _data = self.transform(spec, _data, n_jobs, inverse, as_dataframe)

        if data is None:
            self._processed = _data
            self._formatted = True
            return self
        else:
            return _data

    def inverse(self, data, n_jobs=1, as_dataframe=False):
        _data = self.format(data, n_jobs, True, as_dataframe)
        _data = self.preprocess(_data, n_jobs, True, as_dataframe)
        return _data

    def check_length(self, data):
        length = len(data)
        if isinstance(data, dict):
            length = -1
            for k in data:
                if length < 0:
                    length = len(data[k])
                elif length != len(data[k]):
                    raise Exception("Length of each data column is mismatch.")

        return length

    def iterator(self, data=None, batch_size=32, epoch=-1, n_jobs=1,
                 output_epoch_end=False):

        _data = data if data is not None else self._processed
        data_length = self.check_length(_data)
        steps_per_epoch = data_length // batch_size

        def generator():
            indices = np.arange(data_length)
            count = 0
            _epoch = 0
            while True:
                done = False
                if count > 0 and count % steps_per_epoch == 0:
                    done = True
                    _epoch += 1
                    count = 0
                    if epoch > 0 and _epoch >= epoch:
                        break

                selected = np.random.choice(indices, size=batch_size,
                                            replace=False)

                if isinstance(_data, pd.DataFrame):
                    batch = _data.iloc[selected, :]
                else:
                    batch = {}
                    for key in _data:
                        if isinstance(_data[key], pd.Series):
                            batch[key] = _data[key].iloc[selected]
                            batch[key].reset_index(inplace=True, drop=True)
                        else:
                            batch[key] = _data[key][selected]

                count += 1
                if not self._formatted:
                    batch = self.format(batch, n_jobs=n_jobs)
                if output_epoch_end:
                    batch = [batch, done]
                yield batch

        return steps_per_epoch, generator

    def iterate(self, data=None, batch_size=32, epoch=1, n_jobs=1,
                output_epoch_end=False):
        _, iterator = self.iterator(data, batch_size, epoch, n_jobs,
                                    output_epoch_end)
        for batch in iterator():
            yield batch



from collections import Counter
import numbers
from chariot.util import apply_map
from chariot.transformer.base_preprocessor import BasePreprocessor
from chariot.transformer.tokenizer.token import Token
from chariot.resource.word_vector import WordVector


class Vocabulary(BasePreprocessor):

    def __init__(self, padding="@@PADDING@@", unknown="@@UNKNOWN@@",
                 begin_of_sequence="@@BEGIN_OF_SEQUENCE@@",
                 end_of_sequence="@@END_OF_SEQUENCE@@",
                 max_df=1.0, min_df=1, vocab_size=-1, ignore_blank=True,
                 copy=True):
        super().__init__(copy)
        self._vocab = []
        self._padding = padding
        self._unknown = unknown
        self._begin_of_sequence = begin_of_sequence
        self._end_of_sequence = end_of_sequence
        self.max_df = max_df
        self.min_df = min_df
        self.vocab_size = vocab_size
        self.ignore_blank = ignore_blank

        if max_df < 0 or min_df < 0:
            raise ValueError("Negative value for max_df or min_df")

    @classmethod
    def from_file(cls, path, padding="@@PADDING@@", unknown="@@UNKNOWN@@",
                  begin_of_sequence="", end_of_sequence="",
                  max_df=1.0, min_df=1, vocab_size=-1, ignore_blank=True, copy=True):

        instance = cls(padding, unknown, begin_of_sequence, end_of_sequence,
                       max_df, min_df, vocab_size, ignore_blank, copy)

        with open(path, encoding="utf-8") as f:
            words = f.readlines()
            words = [w.strip() for w in words]
        instance._vocab = words
        return instance

    def set(self, list_or_file):
        reserved = [self._padding, self._unknown,
                    self._begin_of_sequence, self._end_of_sequence]
        reserved = [r for r in reserved if r]

        if isinstance(list_or_file, (list, tuple)):
            def get_surface(token):
                if isinstance(token, Token):
                    return token.surface
                else:
                    return token
            vocab = [get_surface(t) for t in list_or_file]
        else:
            with open(list_or_file, encoding="utf-8") as f:
                words = f.readlines()
                words = [w.strip() for w in words]
            vocab = words

        reserved = [r for r in reserved if r not in vocab]
        if self.ignore_blank:
            vocab = [v for v in vocab if v.strip()]
        vocab = reserved + vocab
        self._vocab = vocab

    def get(self):
        return self._vocab

    @property
    def count(self):
        return len(self._vocab)

    @property
    def unk(self):
        return self.__return_index(self._unknown)

    @property
    def pad(self):
        return self.__return_index(self._padding)

    @property
    def bos(self):
        return self.__return_index(self._begin_of_sequence)

    @property
    def eos(self):
        return self.__return_index(self._end_of_sequence)

    def __return_index(self, word):
        if word in self._vocab:
            return self._vocab.index(word)
        else:
            return -1

    def token_to_words(self, tokens):
        words = [t if isinstance(t, str) else t.surface for t in tokens]
        return [w.strip() for w in words]

    def apply(self, words):
        _words = self.token_to_words(words)
        indices = []
        for w in _words:
            if w in self._vocab:
                indices.append(self._vocab.index(w))
            else:
                indices.append(self.unk)
        return indices

    def inverse(self, indices, exclude_padding=True):
        vocab = self._vocab
        pad = self.pad
        # Make text exclude padding
        words = [vocab[i] for i in indices]
        if exclude_padding:
            words = [vocab[i] for i in indices if i != pad]
        return words

    def transform(self, X):
        if len(self._vocab) == 0:
            raise Exception("Vocabulary has not made yet. Plase execute fit.")
        return super().transform(X)

    def inverse_transform(self, X):
        if len(self._vocab) == 0:
            raise Exception("Vocabulary has not made yet. Plase execute fit.")
        return apply_map(X, self.inverse, inplace=self.copy)

    def fit(self, X, y=None):
        vocab = Counter()
        length = len(X)
        if isinstance(X, dict):
            length = len(list(X.values)[0])

        def update_vocab(element):
            words = self.token_to_words(element)
            if self.ignore_blank:
                words = [w for w in words if w.strip()]
            vocab.update(words)

        apply_map(X, update_vocab)

        reserved = [self._padding, self._unknown,
                    self._begin_of_sequence, self._end_of_sequence]
        reserved = [r for r in reserved if r]  # filter no setting token

        selected = []
        if self.vocab_size > 0:
            for term, count in vocab.most_common():
                if len(selected) < self.vocab_size:
                    selected.append(term)
                else:
                    break
        else:
            min_limit = (self.min_df
                         if isinstance(self.min_df, numbers.Integral)
                         else self.min_df * length)
            max_limit = (self.max_df
                         if isinstance(self.max_df, numbers.Integral)
                         else self.max_df * length)

            for term, count in vocab.most_common():
                if count < min_limit or count > max_limit:
                    continue
                else:
                    selected.append(term)

        reserved = [r for r in reserved if r not in selected]
        self._vocab = reserved + selected
        return self

    def make_embedding(self, word_vector_path,
                       encoding="utf-8", progress=False):
        wv = WordVector(word_vector_path, encoding)
        embedding = wv.load_embedding(self._vocab, progress=progress)
        return embedding



from chariot.transformer.token.base import TokenNormalizer


class BaseFormNormalizer(TokenNormalizer):

    def __init__(self, copy=True):
        super().__init__(copy)

    def apply(self, tokens):
        for t in tokens:
            t.set_surface(t.base_form)

        return tokens



from chariot.transformer.token.base import TokenFilter
from chariot.util import apply_map
from collections import Counter


class WordFrequencyFilter(TokenFilter):

    def __init__(self, n, min_freq=1, copy=True):
        super().__init__(copy)
        self.n = n
        self.min_freq = min_freq
        self._filter_words = []

    def token_to_words(self, tokens):
        words = [t if isinstance(t, str) else t.surface for t in tokens]
        return [w.strip() for w in words]

    def fit(self, X, y=None):
        words_counter = Counter()
        def update_word_counts(element):
            words = self.token_to_words(element)
            words_counter.update(words)

        apply_map(X, update_word_counts)
        common_words = {word for word, freq in words_counter.most_common(self.n)}
        rare_words = {word for word, freq in words_counter.most_common() if freq <= self.min_freq}
        filter_words = common_words.union(rare_words)
        self._filter_words = filter_words
        return self

    def transform(self, X):
        if len(self._filter_words) == 0:
            raise Exception("Filter words has not made yet. Plase execute fit.")
        return super().transform(X)

    def apply(self, tokens):
        if len(tokens) == 0:
            return tokens
        else:
            return [t for t in tokens if t.surface not in self._filter_words]



from .stopword_filter import StopwordFilter
from .base_form_normalizer import BaseFormNormalizer
from .number_normalizer import NumberNormalizer
from .word_frequency_filter import WordFrequencyFilter



from chariot.transformer.tokenizer.token import Token


class SplitTokenizer():

    def __init__(self):
        pass

    def tokenize(self, text):
        return [Token(t, token_type="-") for t in text.split()]



from chariot.transformer.tokenizer.token import Token


class SpacyTokenizer():

    def __init__(self, lang):
        disables = ["textcat", "ner", "parser"]
        import spacy
        if lang in ["en", "de", "es", "pt", "fr", "it", "nl"]:
            nlp = spacy.load(lang, disable=disables)
        else:
            nlp = spacy.load("xx", disable=disables)

        self._nlp = nlp

    def tokenize(self, text):
        tokens = self._nlp(text)
        return [Token(t) for t in tokens]



from sklearn.base import BaseEstimator, TransformerMixin


class BaseGenerator(BaseEstimator, TransformerMixin):

    def __init__(self):
        pass

    def transform(self, data, index, length):
        raise Exception("You should implements transform in subclass.")

    def generate(self, data, index, length):
        raise Exception("You should implements generate in subclass.")



from collections import namedtuple
from chariot.transformer.tokenizer.token import Token


class MeCabTokenizer():
    JanomeToken = namedtuple("JanomeToken", ("surface", "part_of_speech",
                                             "infl_type", "infl_form",
                                             "base_form", "reading",
                                             "phonetic"))

    def __init__(self):
        import MeCab
        self.tagger = MeCab.Tagger("-Ochasen")

    def tokenize(self, text):
        self.tagger.parse("")
        node = self.tagger.parseToNode(text)
        tokens = []
        while node:
            # Ignore BOS/EOS
            if node.surface:
                surface = node.surface
                features = node.feature.split(",")
                if len(features) < 9:
                    pad_size = 9 - len(features)
                    features += ["*"] * pad_size

                token = MeCabTokenizer.JanomeToken(
                            surface, ",".join(features[:4]),
                            features[4], features[5],
                            features[6], features[7],
                            features[8])
                token = Token(token, token_type="ja")
                tokens.append(token)
            node = node.next
        return tokens


class JanomeTokenizer():

    def __init__(self):
        from janome.tokenizer import Tokenizer
        self.tokenizer = Tokenizer()

    def tokenize(self, text):
        tokens = self.tokenizer.tokenize(text)
        tokens = [Token(t, token_type="ja") for t in tokens]
        return tokens



from chariot.transformer.base_preprocessor import BasePreprocessor


class TokenNormalizer(BasePreprocessor):

    def apply(self, tokens):
        raise Exception("You have to implements apply")


class TokenFilter(BasePreprocessor):

    def apply(self, tokens):
        raise Exception("You have to implements apply")



import os
import mmap
from tqdm import tqdm


class DataFile():

    def __init__(self, path, encoding="utf-8"):
        self.path = path
        self.encoding = encoding
        file_name = os.path.basename(path)
        base_name, ext = os.path.splitext(file_name)
        self.base_name = base_name
        self.ext = ext

    def exists(self):
        return os.path.exists(self.path)

    def convert(self, data_dir_to="", add_attribute="",
                attribute_to=(), ext_to=""):
        _dir = os.path.dirname(self.path)
        elements = self._elements()
        ext = self.ext

        if data_dir_to:
            _dir = os.path.join(_dir, "../" + data_dir_to)

        if add_attribute:
            elements.append(add_attribute)
        elif len(attribute_to) > 0:
            # File name format is name + "__".join(attributes)
            # So attribute is elements[1:]
            for a in attribute_to:
                if a in elements[1:]:
                    index = elements[1:].index(a)
                    elements[1 + index] = attribute_to[a]
        if ext_to:
            ext = ext_to

        base_name = "__".join(elements)
        new_path = os.path.join(_dir, base_name + ext)
        return self.__class__(new_path)

    @property
    def name(self):
        return self._elements[0]

    @property
    def attributes(self):
        return self._elements[1:]

    def _elements(self):
        elements = self.base_name.split("__")
        return elements

    def get_line_count(self):
        count = 0
        with open(self.path, "r+") as f:
            buf = mmap.mmap(f.fileno(), 0)
            while buf.readline():
                count += 1
        return count

    def fetch(self, progress=False):
        total_count = 0
        if progress:
            total_count = self.get_line_count()
        with open(self.path, encoding=self.encoding) as f:
            iterator = f
            if progress:
                iterator = tqdm(f, total=total_count)
            for line in iterator:
                yield line.strip()

    def to_array(self):
        lines = []
        with open(self.path, encoding=self.encoding) as f:
            lines = f.readlines()
            lines = [ln.strip() for ln in lines]
        return lines



class Token():

    def __init__(self, token, token_type="spaCy"):
        self._token = token
        self.token_type = token_type
        self.__surface = ""

    def set_surface(self, surface):
        self.__surface = surface

    @property
    def is_spacy(self):
        return self.token_type == "spaCy"

    @property
    def is_ja(self):
        return self.token_type == "ja"

    @property
    def surface(self):
        if self.__surface:
            return self.__surface
        elif self.token_type == "spaCy":
            return self._token.text
        elif self.token_type == "ja":
            return self._token.surface
        else:
            return self._token

    @property
    def base_form(self):
        if self.token_type == "spaCy":
            return self._token.lemma_
        elif self.token_type == "ja":
            return self._token.base_form
        else:
            return "-"

    @property
    def pos(self):
        if self.token_type == "spaCy":
            return self._token.pos_
        elif self.token_type == "ja":
            pos = self._token.part_of_speech.split(",")[0]
            return pos
        else:
            return "-"

    @property
    def tag(self):
        if self.token_type == "spaCy":
            return self._token.tag_
        elif self.token_type == "ja":
            tag = self._token.part_of_speech.split(",")[1]
            return tag
        else:
            return "-"

    def __repr__(self):
        return "<{}:{}>".format(self.surface, self.pos)

    def __reduce_ex__(self, proto):
        return type(self), (self._token, self.token_type)



import numpy as np
from chariot.transformer.generator.base import BaseGenerator


class TargetGenerator(BaseGenerator):

    def __init__(self):
        super().__init__()

    def generate(self, data, index, length):
        _from = data[index:index+length]
        _to = self.transform(data, index, length)
        return _from, _to


class ShiftedTarget(TargetGenerator):

    def __init__(self, shift=1):
        super().__init__()
        self.shift = shift

    def transform(self, data, index, length):
        _to = data[index+self.shift:index+self.shift+length]
        return _to


class ShuffledTarget(TargetGenerator):

    def __init__(self):
        super().__init__()

    def transform(self, data, index, length):
        original = data[index:index+length]
        _to = np.apply_along_axis(np.random.permutation, 1, original)
        return _to



from .unicode_normalizer import UnicodeNormalizer
from .symbol_filter import SymbolFilter
from .lower_normalizer import LowerNormalizer
from .regular_expression_replacer import RegularExpressionReplacer



from chariot.transformer.text.base import TextFilter
import re


class RegularExpressionReplacer(TextFilter):

    def __init__(self, pattern, replacement, copy=True):
        super().__init__(copy)
        self.pattern = pattern
        self.replacement = replacement

    def apply(self, text):
        return re.sub(self.pattern, self.replacement, text)



from chariot.transformer.base_preprocessor import BasePreprocessor
from chariot.util import apply_map
from .ja_tokenizer import MeCabTokenizer
from .ja_tokenizer import JanomeTokenizer
from .spacy_tokenizer import SpacyTokenizer
from .split_tokenizer import SplitTokenizer


class Tokenizer(BasePreprocessor):

    def __init__(self, lang="en", copy=True):
        super().__init__(copy)
        self.lang = lang
        self._tokenizer = None
        self.set_tokenizer()

    def set_tokenizer(self):
        if self.lang == "ja":
            try:
                self.tokenizer = MeCabTokenizer()
            except Exception as ex:
                self.tokenizer = JanomeTokenizer()
        elif self.lang is None:
            self.tokenizer = SplitTokenizer()
        else:
            self.tokenizer = SpacyTokenizer(self.lang)

    def transform(self, X):
        _X = apply_map(X, self.tokenizer.tokenize, self.copy)
        return _X

    def __reduce_ex__(self, proto):
        return type(self), (self.lang, self.copy)



from chariot.transformer.token.base import TokenFilter


class StopwordFilter(TokenFilter):

    def __init__(self, lang, copy=True):
        super().__init__(copy)
        self.lang = lang

    def apply(self, tokens):
        if len(tokens) == 0:
            return tokens
        else:
            sample = tokens[0]
            if sample.is_spacy:
                return [t for t in tokens if not t._token.is_stop]
            elif sample.is_ja:
                return [t for t in tokens if t.surface not in STOP_WORD_JA]
            else:
                return tokens


STOP_WORD_JA = set("""
あそこ あたり あちら あっち あと あな あなた あれ いくつ いつ
 いま いや いろいろ うち おおまか おまえ おれ がい かく かたち
 かやの から がら きた くせ ここ こっち こと ごと こちら
 ごっちゃ これ これら ごろ さまざま さらい さん しかた しよう すか
 ずつ すね すべて ぜんぶ そう そこ そちら そっち そで それ
 それぞれ それなり たくさん たち たび ため だめ ちゃ ちゃん てん
 とおり とき どこ どこか ところ どちら どっか どっち どれ なか
 なかば なに など なん はじめ はず はるか ひと ひとつ ふく
 ぶり べつ へん ぺん ほう ほか まさ まし まとも まま
 みたい みつ みなさん みんな もと もの もん やつ よう よそ
 わけ わたし ハイ 上 中 下 字 年 月 日
 時 分 秒 週 火 水 木 金 土 国
 都 道 府 県 市 区 町 村 各 第
 方 何 的 度 文 者 性 体 人 他
 今 部 課 係 外 類 達 気 室 口
 誰 用 界 会 首 男 女 別 話 私
 屋 店 家 場 等 見 際 観 段 略
 例 系 論 形 間 地 員 線 点 書
 品 力 法 感 作 元 手 数 彼 彼女
 子 内 楽 喜 怒 哀 輪 頃 化 境
 俺 奴 高 校 婦 伸 紀 誌 レ 行
 列 事 士 台 集 様 所 歴 器 名
 情 連 毎 式 簿 回 匹 個 席 束
 歳 目 通 面 円 玉 枚 前 後 左
 右 次 先 春 夏 秋 冬 一 二 三
 四 五 六 七 八 九 十 百 千 万
 億 兆 下記 上記 時間 今回 前回 場合 一つ 年生
 自分 ヶ所 ヵ所 カ所 箇所 ヶ月 ヵ月 カ月 箇月 名前
 本当 確か 時点 全部 関係 近く 方法 我々 違い 多く
 扱い 新た その後 半ば 結局 様々 以前 以後 以降 未満
 以上 以下 幾つ 毎日 自体 向こう 何人 手段 同じ 感じ
""".split())



from chariot.transformer.text.base import TextNormalizer


class LowerNormalizer(TextNormalizer):

    def __init__(self, copy=True):
        super().__init__(copy)

    def apply(self, text):
        return text.lower()



import pandas as pd
import numpy as np
from chariot.transformer.formatter.base import BaseFormatter
from chariot.preprocessor import Preprocessor


class Padding(BaseFormatter):

    def __init__(self, padding=0, length=-1,
                 begin_of_sequence=False, end_of_sequence=False,
                 begin_of_sequence_index=-1, end_of_sequence_index=-1):
        super().__init__()
        self.padding = padding
        self.length = length

        self.begin_of_sequence = begin_of_sequence
        self.end_of_sequence = end_of_sequence
        self._begin_of_sequence = begin_of_sequence_index
        self._end_of_sequence = end_of_sequence_index

        if type(self.begin_of_sequence) != bool:
            self._begin_of_sequence = self.begin_of_sequence
            self.begin_of_sequence = True

        if type(self.end_of_sequence) != bool:
            self._end_of_sequence = self.end_of_sequence
            self.end_of_sequence = True

    def transfer_setting(self, vocabulary_or_preprocessor):
        vocabulary = vocabulary_or_preprocessor
        if isinstance(vocabulary_or_preprocessor, Preprocessor):
            vocabulary = vocabulary_or_preprocessor.vocabulary

        self.padding = vocabulary.pad
        if self.begin_of_sequence:
            self._begin_of_sequence = vocabulary.bos
        if self.end_of_sequence:
            self._end_of_sequence = vocabulary.eos

    def transform(self, column):
        def adjust(sequence, length):
            if self.begin_of_sequence:
                sequence = [self._begin_of_sequence] + sequence
            if self.end_of_sequence:
                sequence = sequence + [self._end_of_sequence]

            if length > 0:
                if len(sequence) < length:
                    pad_size = length - len(sequence)
                    sequence = sequence + [self.padding] * pad_size
                elif len(sequence) > length:
                    sequence = sequence[:length]

            return np.array(sequence)

        length = self.length
        if length < 0:
            length = max([len(seq) for seq in column])

        if isinstance(column, pd.Series):
            applied = column.apply(lambda x: adjust(x, length)).tolist()
            return np.array(applied)
        else:
            return np.array([adjust(x, length) for x in column])

    def inverse_transform(self, column):
        def inverse(sequence):
            excludes = [self.padding,
                        self._begin_of_sequence, self._end_of_sequence]
            inversed = [t for t in sequence if t not in excludes]
            return inversed

        if isinstance(column, pd.Series):
            return column.apply(lambda x: inverse(x))
        else:
            return [inverse(x) for x in column]

    def __reduce_ex__(self, proto):
        return type(self), (self.padding, self.length,
                            self.begin_of_sequence, self.end_of_sequence,
                            self._begin_of_sequence, self._end_of_sequence)



import os
from pathlib import Path
import shutil
import re
from zipfile import ZipFile
import gzip
import pandas as pd


class Storage():

    def __init__(self, root):
        """Provide the access to the data folder.
        The data folder structure should follow the Cookiecutter Data Science
        manner.
        https://drivendata.github.io/cookiecutter-data-science/

        ROOT
         └── data
              ├── external     <- Data from third party sources.
              ├── interim      <- Intermediate data that has been transformed.
              ├── processed    <- The final, canonical datasets for modeling.
              └── raw          <- The original, immutable data dump.

        Args:
            root: path to root directory (parent directory of data, models).
        """
        self.root = root

    @classmethod
    def setup_data_dir(cls, path):
        if not os.path.exists(path):
            raise Exception("{} does not exist".format(path))

        root = Path(path)
        data_root = root.joinpath("data")
        data_root.mkdir(exist_ok=True)
        for _dir in ["raw", "external", "interim", "processed"]:
            data_root.joinpath(_dir).mkdir(exist_ok=True)

        storage = cls(path)
        return storage

    def data_path(self, target=""):
        return os.path.join(self.root, "data/{}".format(target))

    def read(self, target, encoding="utf-8", delimiter=",",
             header="infer", names=None):
        path = self.data_path(target)
        if names is not None:
            df = pd.read_csv(path, delimiter=delimiter, header=None,
                             names=names)
        else:
            df = pd.read_csv(path, delimiter=delimiter, header=header)
        return df

    def chazutsu(self, path, columns=None, target="",
                 separator="\t", pattern=()):

        from chazutsu.datasets.framework.resource import Resource
        r = Resource(path, columns, target, separator, pattern)
        return r

    def chakin(self, lang="", number=-1, name=""):
        import chakin
        if lang:
            chakin.search(lang)
        elif number > -1 or name:
            path = self.data_path("external")
            if not os.path.exists(path):
                os.mkdir(path)

            table = chakin.downloader.load_datasets()

            index = number
            if number < 0:
                index = table.index[table["Name"] == name].tolist()
                index = index[0]

            _name = table.iloc[index]["Name"].lower()

            for ext in [".txt", ".vec"]:
                check_path = os.path.join(path, _name) + ext
                if os.path.exists(check_path):
                    return check_path

            vec_path = chakin.download(index, path)

            base, ext = os.path.splitext(vec_path)
            _dir = os.path.dirname(vec_path)
            if ext == ".vec":
                vec_path = os.rename(vec_path, os.path.join(_dir, _name + ext))
            elif ext in [".zip", ".gz"]:
                _path = self.expand(vec_path, ext)
                os.remove(vec_path)
                vec_path = _path

            return vec_path

        else:
            raise Exception("You have to specify lang to search or "
                            "number/name to download")

    """
    def _to_snake(self, name):
        _name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        _name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", _name).lower()
        return _name
    """

    def expand(self, path, ext):
        location = Path(os.path.dirname(path))
        file_name = os.path.basename(path)
        file_root, ext = os.path.splitext(file_name)

        if ext == ".gz":
            file_path = location.joinpath(file_root)
            with gzip.open(path, "rb") as f_in:
                with open(file_path, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
            return file_path
        else:
            dir_path = os.path.join(location, file_root)
            if os.path.exists(dir_path):
                print("The file {} is already expanded".format(
                        os.path.abspath(path)))

            if ext == ".zip":
                with ZipFile(path) as zip:
                    zip.extractall(location)

            return dir_path



import numpy as np
from chariot.resource.data_file import DataFile


class WordVector(DataFile):

    def __init__(self, path, encoding="utf-8"):
        super().__init__(path, encoding)

    def load_embedding(self, vocab, progress=False):
        embedding = None

        initialized = False
        for line in self.fetch(progress):
            values = line.split()
            word = values[0]
            vector = values[1:]
            if not initialized:
                if word.isdigit() and\
                   (len(vector) == 1 and vector[0].isdigit()):
                    # Word2Vec format: First row is vocab_size & vector_size
                    # https://github.com/3Top/word2vec-api/issues/6#issuecomment-179339511
                    embedding_size = int(vector[0])
                    embedding = np.zeros((len(vocab), embedding_size))
                    initialized = True
                else:
                    embedding_size = len(vector)
                    embedding = np.zeros((len(vocab), embedding_size))
                    initialized = True

            if word in vocab:
                index = vocab.index(word)
                vector = np.asarray(values[1:], dtype="float32")
                embedding[index] = vector
        return embedding

    def load(self, progress=False):
        key_vector = {}
        vector_size = -1
        for line in self.fetch(progress):
            values = line.split()
            word = values[0]
            vector = values[1:]
            if vector_size < 0:
                if word.isdigit() and\
                   (len(vector) == 1 and vector[0].isdigit()):
                    vector_size = int(vector[0])
                else:
                    vector_size = len(vector)

            vector = np.asarray(values[1:], dtype="float32")
            key_vector[word] = vector
        return key_vector

    def load_model(self, binary):
        from gensim.models import KeyedVectors
        model = KeyedVectors.load_word2vec_format(self.path, binary=binary)
        return model



import re
import copy
import numpy as np
from sklearn.externals import joblib
from sklearn.utils.metaestimators import _BaseComposition
from sklearn.base import BaseEstimator, TransformerMixin
from chariot.transformer.text.base import TextNormalizer, TextFilter
from chariot.transformer.token.base import TokenFilter, TokenNormalizer
from chariot.transformer.vocabulary import Vocabulary
from chariot.transformer.tokenizer import Tokenizer


class Preprocessor(_BaseComposition, BaseEstimator, TransformerMixin):

    def __init__(self, tokenizer=None,
                 text_transformers=(), token_transformers=(),
                 vocabulary=None, other_transformers=()):
        self.tokenizer = tokenizer
        if isinstance(self.tokenizer, str):
            self.tokenizer = Tokenizer(self.tokenizer)
        self.text_transformers = list(text_transformers)
        self.token_transformers = list(token_transformers)
        self.vocabulary = vocabulary
        self.other_transformers = list(other_transformers)

    def stack(self, transformer):
        if isinstance(transformer, Tokenizer):
            self.tokenizer = transformer
        elif isinstance(transformer, (TextNormalizer, TextFilter)):
            self.text_transformers.append(transformer)
        elif isinstance(transformer, (TokenFilter, TokenNormalizer)):
            self.token_transformers.append(transformer)
        elif isinstance(transformer, Vocabulary):
            self.vocabulary = transformer
        elif isinstance(transformer, (BaseEstimator, TransformerMixin)):
            self.other_transformers.append(transformer)
        else:
            raise Exception("Can't append transformer to the Preprocessor")
        return self

    def _to_snake(self, name):
        _name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
        _name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", _name).lower()
        return _name

    @property
    def _transformers(self):
        transformers = list(self.text_transformers)
        if self.tokenizer:
            transformers += [self.tokenizer]
        transformers += self.token_transformers
        if self.vocabulary:
            transformers += [self.vocabulary]
        transformers += self.other_transformers

        return (
            (self._to_snake(t.__class__.__name__) + "_at_{}".format(i), t)
            for i, t in enumerate(transformers)
        )

    def _validate_transformers(self):
        names, transformers = zip(*self._transformers)

        # validate names
        self._validate_names(names)

        # validate estimators
        for t in transformers:
            if t is None:
                continue
            if (not (hasattr(t, "fit") or hasattr(t, "fit_transform")) or not
                    hasattr(t, "transform")):
                raise TypeError("All transformer should implement fit and "
                                "transform. '%s' (type %s) doesn't" %
                                (t, type(t)))

    def transform(self, X):
        self._validate_transformers()
        Xt = self.check_array(X, True)
        for name, t in self._transformers:
            Xt = t.transform(Xt)

        return Xt

    def inverse_transform(self, X):
        self._validate_transformers()
        Xt = X
        for name, t in list(self._transformers)[::-1]:
            if hasattr(t, "inverse_transform"):
                Xt = t.inverse_transform(Xt)
        return Xt

    def fit(self, X, y=None):
        self._validate_transformers()
        Xt = X
        copied = False
        for name, t in self._transformers:
            t.fit(Xt)
            if not isinstance(t, Vocabulary):
                original_copy_setting = t.copy
                if not copied:
                    # Do not transform original X
                    t.copy = False
                    copied = True
                else:
                    t.copy = False
                Xt = t.transform(Xt)
                t.copy = original_copy_setting
        return self

    def fit_transform(self, X, y=None):
        self._validate_transformers()
        Xt = X
        for name, t in self._transformers:
            Xt = t.fit_transform(Xt)
        return Xt

    def check_array(self, X, copy_obj=True):
        if not copy_obj:
            return X
        else:
            if isinstance(X, (np.ndarray, np.generic)):
                return np.array(X)
            else:
                return copy.deepcopy(X)

    def save(self, path):
        joblib.dump(self, path)

    @classmethod
    def load(cls, path):
        instance = joblib.load(path)
        return instance



import itertools
import numpy as np
import pandas as pd
from chariot.dataset_preprocessor import DatasetPreprocessor
from chariot.dataset_preprocessor import ProcessBuilder
from chariot.transformer.formatter.base import BaseFormatter
from chariot.transformer.generator.base import BaseGenerator
from chariot.transformer.generator.target_generator import TargetGenerator


class LanguageModelProcessBuilder(ProcessBuilder):

    def __init__(self, dp, name):
        super().__init__(dp, name)

    def by(self, processor, reference=None):
        if isinstance(processor, BaseGenerator):
            self.dp.generator[self.key] = processor
        else:
            super().by(processor, reference)
        return self


class SentenceToBatchTransformer(BaseFormatter):

    def __init__(self, batch_size):
        self.batch_size = batch_size

    def transform(self, column):
        sentences = column
        if isinstance(column, pd.Series):
            sentences = column.values
        elif isinstance(column, (list, tuple)):
            sentences = np.array(column)

        if len(column.shape) > 1 and column.shape[1] == 1:
            sentences = column.flatten()

        adjusted = np.array(list(itertools.chain.from_iterable(sentences)))
        limit = (len(adjusted) // self.batch_size) * self.batch_size
        # None x batch_size matrix
        adjusted = adjusted[:limit].reshape((self.batch_size, -1)).T
        return adjusted


class LanguageModelPreprocessor(DatasetPreprocessor):

    def __init__(self, spec=()):
        super().__init__(spec)
        self.generator = {}

    def process(self, name):
        return LanguageModelProcessBuilder(self, name)

    def generate(self, column, index=0, sequence_length=12):
        if len(self.generator) == 0:
            raise Exception("Language model should have at least one generator")

        generated_pairs = []
        for k in self.generator:
            from_field, to_field = k
            g = self.generator[k]
            left, right = g.generate(column, index, sequence_length)

            if isinstance(g, TargetGenerator):
                generated_pairs.append((left, right))
            else:
                generated_pairs.append((right, left))

        return generated_pairs

    def iterator(self, data=None, batch_size=32, sequence_length=12,
                 epoch=-1, sequencial=True, output_epoch_end=False,
                 n_jobs=1):

        _data = data if data is not None else self._processed
        for key in self.generator:
            _from, _to = key

        batch_all = SentenceToBatchTransformer(batch_size).transform(_data[_from])
        steps_per_epoch = (len(batch_all) - 1) // sequence_length

        def generator():
            count = 0
            _epoch = 0
            index = 0
            while True:
                done = False
                if count > 0 and count % steps_per_epoch == 0:
                    done = True
                    _epoch += 1
                    count = 0
                    index = 0
                    if epoch > 0 and _epoch >= epoch:
                        break

                pairs = self.generate(batch_all, index, sequence_length)

                if sequencial:
                    batch = pairs
                else:
                    # to (batch x sequence x word_id) shape
                    batch = []
                    for f_, t_ in pairs:
                        batch.append((f_.T, np.expand_dims(t_.T, axis=-1)))

                count += 1
                index = index + sequence_length

                if len(self.generator) > 1:
                    if output_epoch_end:
                        batch = [batch, done]
                else:
                    batch = list(batch[0])
                    if output_epoch_end:
                        batch += [done]

                yield batch

        return steps_per_epoch, generator

    def iterate(self, data=None, batch_size=32, sequence_length=12,
                epoch=1, sequencial=True, output_epoch_end=False,
                n_jobs=1):
        _, generator = self.iterator(data, batch_size, sequence_length,
                                     epoch, sequencial, output_epoch_end,
                                     n_jobs)
        for batch in generator():
            yield batch



import pandas as pd
import numpy as np


def apply_map(data, func, inplace=False):

    if isinstance(data, pd.DataFrame):
        # apply and apply_map call function twice.
        result = data if inplace else {}
        for c in data.columns:
            s = list(map(func, data[c].values))
            if inplace:
                result.loc[:, c] = s
            else:
                result[c] = s
        if not inplace:
            result = pd.DataFrame.from_dict(result)
        return result
    elif isinstance(data, pd.Series):
        if inplace:
            for i, e in enumerate(map(func, data.values)):
                data.iat[i] = e
            return data
        else:
            result = pd.Series(list(map(func, data.values)))
            return result
    elif isinstance(data, np.ndarray):
        def _apply_map(x):
            return apply_map(x, func, inplace=inplace)

        if len(data.shape) > 1:
            result = np.apply_along_axis(_apply_map, -1, data)
            return result
        else:
            _data = pd.Series(data)
            return apply_map(_data, func, inplace=inplace)

    elif isinstance(data, dict):
        result = data if inplace else {}
        for k in list(data.keys()):
            result[k] = [func(row) for row in data[k]]
        return result
    elif isinstance(data, (list, tuple)):
        result = data if inplace else ([0] * len(data))
        for i, row in enumerate(data):
            if isinstance(row, (tuple, list)) and len(row) > 0\
                 and isinstance(row[0], (tuple, list)):
                # multi-column list
                result[i] = [func(c) for c in row]
            else:
                result[i] = func(row)
        return result
    else:
        return func(data)

    return data



import os
import shutil
import json
from ast import literal_eval
from pathlib import Path
import tarfile
import numpy as np
import pandas as pd
from sklearn.externals import joblib
from sklearn.base import BaseEstimator, TransformerMixin
from joblib import Parallel, delayed


def _apply(data, target, process, inverse=False):
    if isinstance(data, pd.Series) and data.dtype != "object":
        _data = data.values.reshape((-1, 1))
    elif isinstance(data, np.ndarray) and len(data.shape) < 2:
        _data = data.reshape((-1, 1))
    else:
        _data = data

    if not inverse:
        transformed = process.transform(_data)
    else:
        if hasattr(process, "inverse_transform"):
            transformed = process.inverse_transform(_data)

    return (target, transformed)


class BaseDatasetPreprocessor():

    def __init__(self, spec=()):
        self.spec = spec
        if len(self.spec) == 0:
            self.spec = {}

    def transform(self, spec, data=None, n_jobs=1,
                  inverse=False, as_dataframe=False):
        tasks = []
        keys = []
        if isinstance(data, pd.DataFrame):
            keys = data.columns.tolist()
        else:
            keys = list(data.keys())

        for k in spec:
            from_, to_ = k
            if spec[k] is not None:
                tasks.append((from_, to_, spec[k]))
            if from_ in keys:
                keys.remove(from_)

        transformed = Parallel(n_jobs=n_jobs)(
            delayed(_apply)(data[s], t, p, inverse) for s, t, p in tasks)
        transformed = dict(transformed)

        for k in keys:
            if k not in transformed:
                transformed[k] = data[k]

        if not as_dataframe:
            return transformed
        else:
            applied = pd.DataFrame.from_dict(transformed)
            return applied

    @classmethod
    def _apply_func(cls, d, criteria, func, path=()):
        applied = {}
        for k in d:
            p = list(path) + [k]
            if isinstance(d[k], dict):
                applied[k] = cls._apply_func(d[k], criteria, func, p)
            elif criteria(d[k]):
                applied[k] = func(d[k], p)
            else:
                applied[k] = d[k]

        return applied

    @classmethod
    def _make_file_name(cls, path):
        file_name = ""
        for p in path:
            if isinstance(p, tuple):
                file_name += "-".join(p)
            else:
                file_name += "_" + p
        file_name += ".pkl"
        return file_name

    def save(self, path):
        if os.path.exists(path):
            os.remove(path)
        basename = os.path.basename(path).split(".")[0]
        root = Path(os.path.dirname(path)).joinpath(basename)
        root.mkdir()

        def criteria(item):
            if isinstance(item, (BaseEstimator, TransformerMixin)):
                return True
            else:
                return False

        def func(item, path):
            file_name = self._make_file_name(path)
            joblib.dump(item, root.joinpath(file_name))
            return file_name

        spec = self._apply_func(self.spec, criteria, func)
        with root.joinpath("spec.json").open(mode="w", encoding="utf-8") as f:
            spec = {str(k): v for k, v in spec.items()}
            json.dump(spec, f)

        with tarfile.open(path, "w:gz") as tar:
            tar.add(root, arcname=basename)

        shutil.rmtree(root)

    @classmethod
    def load(cls, path):
        basename = os.path.basename(path).split(".")[0]

        with tarfile.open(path, "r:gz") as tar:
            spec_f = tar.extractfile(basename + "/spec.json")
            spec = json.load(spec_f)
            spec = {literal_eval(k): v for k, v in spec.items()}

            def criteria(item):
                if isinstance(item, str) and item.endswith(".pkl"):
                    return True
                else:
                    return False

            def func(item, path):
                file_name = cls._make_file_name(path)
                file_name = basename + "/" + file_name
                p = joblib.load(tar.extractfile(file_name))
                return p

            _spec = cls._apply_func(spec, criteria, func)

        instance = cls(_spec)
        return instance



from setuptools import setup


readme = ""
with open("README.md", encoding="utf-8") as f:
    readme = f.read()

setup(
    name="chariot",
    version="0.5.2",
    description="Deliver the ready-to-train data to your NLP model.",
    long_description=readme,
    long_description_content_type="text/markdown",
    keywords=["machine learning", "nlp", "natural language processing"],
    author="icoxfog417",
    author_email="icoxfog417@yahoo.co.jp",
    license="Apache License 2.0",
    packages=[
        "chariot",
        "chariot.resource",
        "chariot.transformer",
        "chariot.transformer.formatter",
        "chariot.transformer.generator",
        "chariot.transformer.text",
        "chariot.transformer.token",
        "chariot.transformer.tokenizer"
        ],
    url="https://github.com/chakki-works/chariot",
    install_requires=[
        "numpy>=1.14.4",
        "scipy>=1.1.0",
        "scikit-learn>=0.19.1",
        "spacy==1.9.0",
        "pandas>=0.23.1",
        "chazutsu>=0.8.2",
        "chakin>=0.0.6",
        "tqdm>=4.23.4"
    ],
)



from chariot.transformer.token.base import TokenNormalizer


class NumberNormalizer(TokenNormalizer):

    def __init__(self, digit_character="0", copy=True):
        super().__init__(copy)
        self.digit_character = digit_character

    def apply(self, tokens):
        for t in tokens:
            is_number = False
            if t.is_spacy:
                if t._token.is_digit:
                    t.set_surface(self.digit_character)
                    is_number = True
            elif t.is_ja:
                # print(t.pos, t.tag)
                if (t.pos == "名詞" and t.tag == "数"):
                    t.set_surface(self.digit_character)
                    is_number = True

            escaped = t.surface.replace(".", "").replace(",", "")
            if not is_number and escaped.isdigit():
                t.set_surface(self.digit_character)

        return tokens



from .tokenizer import Tokenizer
from .vocabulary import Vocabulary
# dummy import to call from chariot.transformer module
from .formatter.base import BaseFormatter
from .text.base import BasePreprocessor
from .token.base import BasePreprocessor
from .generator.base import BaseGenerator
